package com.hikcreate.bigdata.hdip.source;

import com.google.gson.Gson;
import org.apache.commons.io.IOUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.FileInputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentLinkedQueue;

/**
 * 采集Oracle表的管理类
 *
 * @author Administrator
 * @Time 2020-08-20 12:12:12
 */

public class OracleSourceService implements QueryCallback {
    /**
     * 用于日志打印
     */
    private static final Logger log = LoggerFactory.getLogger(OracleSourceService.class);
    Map<String, QueryTask> queryTaskMap; //任务集合（线程类）
    private String offsetPath;//offset保存到文件的路径
    private String status = "shutdown";
    private ConcurrentLinkedQueue<QueryTask> queryTasks;
    private Scheduler scheduler;
    private Executor executor;


    /**
     * 方法描述 初始话采集表的信息
     * 处理逻辑 初始化
     *
     * @param
     * @return
     * @throws
     * @author chencong13
     * @Time 2020/8/20 10:30
     */
    private void init(String configPath) throws IOException, SQLException {
        log.info("init ...");
        queryTaskMap = new HashMap<>();
        queryTasks = new ConcurrentLinkedQueue<>();
        String configJson = IOUtils.toString(new FileInputStream(configPath), "utf-8");
        Gson gson = new Gson();
        ConfigBean queryConfig = gson.fromJson(configJson, ConfigBean.class);
        String jdbcType = queryConfig.getJdbcType();
        DBDriver dbDriver = DBDriver.valueOf(jdbcType);
        String driver = dbDriver.getDriver();
        String url = dbDriver.getUrl(queryConfig.getJdbcIp(), queryConfig.getJdbcPort(), queryConfig.getSid());
        DBUtils.initDataSource(url, driver, queryConfig.getJdbcUser(), queryConfig.getJdbcPwd());
        List<TableBean> queryTables = queryConfig.getQueryTables();
        offsetPath = queryConfig.getOffsetPath();
        Map<String, String> offsetMap = new HashMap<>();
        Map<String, Long> fileNoMap = new HashMap<>();
        readOffset(offsetPath, offsetMap, fileNoMap);
        for (TableBean tb : queryTables) {
            String key = tb.getDatabase() + "." + tb.getTable();
            String offset = offsetMap.get(key);
            Long fileNo = fileNoMap.get(key);
            if (offset == null && fileNo == null) {
                offset = tb.getInitOffset();
                fileNo = 0L;
            }
            QueryTask queryTask = new QueryTask(tb, queryConfig.getTempPath(), offset, fileNo, queryConfig.getFtpPath(), jdbcType);
            queryTask.addLister(this);
            queryTaskMap.put(key, queryTask);
        }
        log.info("query table size is {}", queryTaskMap.size());
    }

    /**
     * 方法描述 开始采集
     * 处理逻辑 启动线程开始任务
     *
     * @param
     * @return
     * @throws
     * @author chencong13
     * @Time 2020/8/20 10:30
     */
    public String start(String configPath) {
        if ("started".equals(status)) {
            return "already started";
        }
        try {
            if (queryTaskMap == null) {
                init(configPath);
            }
        } catch (IOException | SQLException e) {
            log.error("oracleSourceService", e);
        }
        log.info("start query...");
        status = "started";
        scheduler = new Scheduler(queryTaskMap, queryTasks);
        executor = new Executor(queryTasks);
        new Thread(scheduler).start();
        new Thread(executor).start();
        return "success";
    }


    /**
     * 方法描述 从文件读取offset
     * 处理逻辑
     *
     * @param path offset的文件路径
     * @return offset的信息
     * @throws
     * @author chencong13
     * @Time 2020/8/20 10:31
     */
    private void readOffset(String path, Map<String, String> offset, Map<String, Long> fileNo) {
        try {
            List<String> strings = IOUtils.readLines(new FileReader(path));
            for (String s : strings) {
                if (CommonUtils.isBlank(s)) {
                    continue;
                }
                String[] split = s.split("=");
                String[] offsets = split[1].split("#");
                offset.put(split[0], offsets[0]);
                fileNo.put(split[0], Long.parseLong(offsets[1]));
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * 方法描述 保存表的offset
     * 处理逻辑
     *
     * @return
     * @throws
     * @author chencong13
     * @Time 2020/8/20 14:34
     */
    public String saveOffset() {
        String msg = "success";
        String path = offsetPath;
        CommonUtils.clearFileContent(offsetPath);
        List<String> temp = new ArrayList<>();
        for (String k : queryTaskMap.keySet()) {
            QueryTask queryTask = queryTaskMap.get(k);
            String offset = queryTask.getOffset();
            if ("-1".equals(offset)) {
                continue;
            }
            long fileNo = queryTask.getFileNo();
            temp.add(k + "=" + queryTask.getOffset() + "#" + fileNo + "\n");
        }
        if (temp.size() == 0) {
            return msg;
        }
        FileWriter fileWriter = null;
        try {
            fileWriter = new FileWriter(path);
            for (String line : temp) {
                fileWriter.write(line);
            }
            fileWriter.flush();
        } catch (IOException e) {
            log.error("error", e);
            throw new IllegalArgumentException("save offset failed");
        } finally {
            try {
                fileWriter.close();
            } catch (IOException e) {
                log.error("error", e);
            }
        }
        return msg;
    }

    /**
     * 方法描述 获取task的当前的offset
     * 处理逻辑
     *
     * @param
     * @return
     * @author chencong13
     * @Time 2020/8/20 15:31
     */
    public String getOffset() {
        StringBuffer sb = new StringBuffer();
        for (String k : queryTaskMap.keySet()) {
            QueryTask queryTask = queryTaskMap.get(k);
            sb.append(k + " = " + queryTask.toString() + " , ");
        }
        sb.append("count = " + queryTaskMap.size());
        String temp = sb.toString();
        sb.setLength(0);
        return temp;
    }

    /**
     * 方法描述 停止掉所有的任务
     * 处理逻辑
     *
     * @param
     * @return
     * @throws
     * @author chencong13
     * @Time 2020/8/20 10:33
     */
    public String shuntDown() {
        log.info("shunt down all ... ");
        if ("shutdown".equals(status)) {
            return "already shutdown";
        }
        saveOffset();
        scheduler.stop();
        executor.stop();
        status = "shutdown";
        for (String key : queryTaskMap.keySet()) {
            QueryTask queryTask = queryTaskMap.get(key);
            queryTask.stop();
        }
        queryTaskMap.clear();
        DruidUtils.closeDruid();
        return "sucess";
    }

    /**
     * 方法描述 回调方法
     * 处理逻辑 保存offset,在任何一个任务出错，或者文件满写50次，都会触发offset的保存
     *
     * @param
     * @return
     * @throws
     * @author chencong13
     * @Time 2020/8/20 11:49
     */
    @Override
    synchronized public void handlerMsg(int waht, String msg) {
        saveOffset();
        switch (waht) {
            case QueryTask.OFFSET_50:
                log.info("50 count trigger save offset :{}", msg);
                break;
            case QueryTask.OFFSET_ERR:
                log.info("error trigger save offset :{},need check", msg);
                queryTaskMap.remove(msg);
                break;
        }
    }

    /**
     * 方法描述 获取服务的当前状态
     * 处理逻辑
     *
     * @param
     * @return
     * @throws
     * @author chencong13
     * @Time 2020/8/21 15:42
     */
    public String getStatus() {
        return status;
    }


}
